/****************************************************************************

 ■作例①：ミニAM/FMラジオ (RDS5830 Module)
		Copyright(C) 2019 電子うさぎ https://電子うさぎ.com/archives/14953

****************************************************************************/

// PIC16F1705 Configuration Bit Settings

// CONFIG1
#pragma config FOSC = INTOSC    // Oscillator Selection Bits (INTOSC oscillator: I/O function on CLKIN pin)
#pragma config WDTE = SWDTEN    // Watchdog Timer Enable (WDT controlled by the SWDTEN bit in the WDTCON register)
#pragma config PWRTE = OFF      // Power-up Timer Enable (PWRT disabled)
#pragma config MCLRE = ON       // MCLR Pin Function Select (MCLR/VPP pin function is MCLR)
#pragma config CP = OFF         // Flash Program Memory Code Protection (Program memory code protection is disabled)
#pragma config BOREN = ON       // Brown-out Reset Enable (Brown-out Reset enabled)
#pragma config CLKOUTEN = OFF   // Clock Out Enable (CLKOUT function is disabled. I/O or oscillator function on the CLKOUT pin)
#pragma config IESO = ON        // Internal/External Switchover Mode (Internal/External Switchover Mode is enabled)
#pragma config FCMEN = ON       // Fail-Safe Clock Monitor Enable (Fail-Safe Clock Monitor is enabled)

// CONFIG2
#pragma config WRT = OFF        // Flash Memory Self-Write Protection (Write protection off)
#pragma config PPS1WAY = ON     // Peripheral Pin Select one-way control (The PPSLOCK bit cannot be cleared once it is set by software)
#pragma config ZCDDIS = ON      // Zero-cross detect disable (Zero-cross detect circuit is disabled at POR)
#pragma config PLLEN = OFF      // Phase Lock Loop enable (4x PLL is enabled when software sets the SPLLEN bit)
#pragma config STVREN = ON      // Stack Overflow/Underflow Reset Enable (Stack Overflow or Underflow will cause a Reset)
#pragma config BORV = HI        // Brown-out Reset Voltage Selection (Brown-out Reset Voltage (Vbor), high trip point selected.)
#pragma config LPBOR = ON       // Low-Power Brown Out Reset (Low-Power BOR is enabled)
#pragma config LVP = ON         // Low-Voltage Programming Enable (Low-voltage programming enabled)

#include <xc.h>
#include "typedefs.h"

//=================================================================

#define ADDR_RDA5830	0b0010001

#define PORT_LED		LATC5
#define PORT_SD			LATC4
#define PORT_PH			RA2

#define KEY_NONE		0x00
#define KEY_PREV		0x01
#define KEY_NEXT		0x02
#define KEY_BAND		0x03
#define KEY_SEEK_PREV	0x11
#define KEY_SEEK_NEXT	0x12
#define KEY_POWER		0x04

#define CHAN_NUM_FM		(sizeof(s_wFreqsFM) / sizeof(WORD))
#define CHAN_NUM_AM		(sizeof(s_wFreqsAM) / sizeof(WORD))
#define MAX_VOLUME		15

//=================================================================

static void InitSystem(void);
static void InitApp(void);
static void Run(void);
static void Normal(void);
static void Standby(void);

static BYTE GetKey(void);
static BYTE GetKeySub(void);
static WORD GetSWV(void);
static void SwitchBand(void);
static void PrevFreq(void);
static void NextFreq(void);
static WORD GetPrevFreq(WORD* pFreq, INT8 nCount, WORD wCur);
static WORD GetNextFreq(WORD* pFreq, INT8 nCount, WORD wCur);
static void Seek(BOOL bUp);

static void EnableRecv(BOOL bEnable);
static void Stereo(BOOL bStereo);
static void SetBand(BOOL bFM);
static void SetFreqFM(WORD wFreq);
static void SetFreqAM(WORD wFreq);
static WORD SeekFM(BOOL bUp);
static WORD SeekAM(BOOL bUp);
static void SetVolume(BYTE bVol);
static void SetMute(BYTE bMute);

static void WriteI2C(BYTE nReg, WORD wData);
static WORD ReadI2C(BYTE nReg);

static void SleepDelay(WORD uWdtPs);

//=================================================================

static WORD s_wFreqsFM[] = { 765, 789, 802, 851, 881, 886, 906, 911, 919, 926, 933, 971, 1065 };
static WORD s_wFreqsAM[] = { 558, 666, 828, 1008, 1143, 1179, 1314 };

static BOOL s_bBand;
static WORD s_wFreqFM;
static WORD s_wFreqAM;
static BYTE s_nVolume;
static BOOL s_bPH;

//=================================================================

// ●メインエントリ
void main(void)
{
	InitSystem();
	InitApp();
	Run();
}

// ●システム初期化
static void InitSystem(void)
{
	// クロック
	OSCCONbits.IRCF = 0b0000; // 31 kHz LF

	// ポート初期化
	PORTA = PORTC = 0x00;
	LATA = LATC = 0x00;
	nWPUEN = 0;
	
	// PORT_LED(RC5)
    ANSELCbits.ANSC5 = 0;
	TRISCbits.TRISC5 = 0;

	// PORT_SD(RC4)
	PORT_SD = 1;
    ANSELCbits.ANSC4 = 0;
	TRISCbits.TRISC4 = 0;

	// PORT_PH(RA2)
    ANSELAbits.ANSA2 = 0;
	TRISAbits.TRISA2 = 1;
	WPUAbits.WPUA2 = 1;

	// SW(RC3)
	TRISCbits.TRISC3 = 1;
	WPUCbits.WPUC3 = 0;

	// Q3(RA5)
	TRISAbits.TRISA5 = 1;
	WPUAbits.WPUA5 = 0;

	// I2C(Master mode)
	RC1PPS = 0b10001; // Rxy source is SDA
	RC0PPS = 0b10000; // Rxy source is SCK/SCL

	ANSELCbits.ANSC1 = 0; // RC1/SDA1(9pin)
	ANSELCbits.ANSC0 = 0; // RC0/SCL1(10pin)

    SSP1ADD = 0x00; // 500÷4=125KHz
	SSP1STAT= 0b00000000;

    SSP1CON1bits.SSPM = 0b1000; // I2C Master mode
    SSP1CON1bits.SSPEN = 1;
	
	// ADC
	ADCON1bits.ADPREF = 0; // VREF+ is connected to VDD
	ADCON1bits.ADNREF = 0; // VREF- is connected to AVSS
	ADCON1bits.ADCS = 0b111; // dedicated RC oscillator
	ADCON1bits.ADFM = 1; // Right justified. 
	ADCON0bits.CHS = 0b00111; // AN7
	ADCON0bits.ADON = 1;

	// 割込み許可
	PEIE = 1;
	GIE = 1;

	SleepDelay(0b00111); // 128ms
}

// ●アプリ初期化
static void InitApp(void)
{
	s_bBand = TRUE;
	s_wFreqFM = s_wFreqsFM[0];
	s_wFreqAM = s_wFreqsAM[0];
	s_nVolume = MAX_VOLUME;
	s_bPH = FALSE;
}

// ●実行ループ
static void Run(void)
{
	for(;;)
	{
		Normal();
		Standby();
	}
}

// ●起動状態
static void Normal(void)
{
	PORT_LED = 1;
	ADCON0bits.ADON = 1;

	// 受信動作開始
	TRISAbits.TRISA5 = 0; // FET(Q3) ON
	SleepDelay(0b00110); // 64ms

	s_bPH = (!PORT_PH);
	EnableRecv(TRUE);
	PORT_SD = s_bPH;
	
	for(;;)
	{
		BYTE nKey = GetKey();
		switch(nKey)
		{
		default:
			// 操作なし
			if (PORT_PH == s_bPH)
			{
				// ヘッドホン⇔スピーカー
				s_bPH = (!PORT_PH);

				Stereo(s_bPH);
				SetVolume((s_bPH)? s_nVolume : MAX_VOLUME);

				PORT_SD = s_bPH;
			}
			break;

		case KEY_BAND:
			SwitchBand();
			break;

		case KEY_PREV:
			if (!s_bPH)
			{
				PrevFreq();
			}
			else
			{
				if (0 < s_nVolume)
					SetVolume(--s_nVolume);
			}
			break;

		case KEY_NEXT:
			if (!s_bPH)
			{
				NextFreq();
			}
			else
			{
				if (s_nVolume < MAX_VOLUME)
					SetVolume(++s_nVolume);
			}
			break;

		case KEY_SEEK_PREV:
			if (!s_bPH)
			{
				Seek(FALSE);
			}
			else
			{
				PrevFreq();
			}
			while(GetKey() != KEY_NONE)
				SleepDelay(0b00011); // 8ms
			break;

		case KEY_SEEK_NEXT:
			if (!s_bPH)
			{
				Seek(TRUE);
			}
			else
			{
				NextFreq();
			}
			while(GetKey() != KEY_NONE)
				SleepDelay(0b00011); // 8ms
			break;

		case KEY_POWER:
			return;
		}
	}
}

// ●スタンバイ
static void Standby(void)
{
	PORT_LED = 0;

	// 受信動作終了
	PORT_SD = 1;
	EnableRecv(FALSE);
	TRISAbits.TRISA5 = 1; // FET(Q3) OFF
	
	// 電源SW押下待ち
	ADCON0bits.ADON = 0;

	// SW(RC3)→Digital
    ANSELCbits.ANSC3 = 0;
	
	IOCCPbits.IOCCP3 = 1;
	IOCIE = 1;
	
	SLEEP();

	IOCIE = 0;
	IOCCPbits.IOCCP3 = 0;

	// SW(RC3)→Analog
    ANSELCbits.ANSC3 = 1;
}

//-----------------------------------------------------------------

// ●キー入力
static BYTE GetKey(void)
{
	BYTE nKey1 = GetKeySub();
	if (nKey1 == KEY_NONE)
		return KEY_NONE;
	
	int i;
	for(i = 0; i < 35; i++)
	{
		BYTE nKey2 = GetKeySub();
		if (nKey1 == nKey2)
			continue;
		
		return (i > 2)? nKey1 : KEY_NONE;
	}
	
	if (nKey1 == KEY_PREV)
		return KEY_SEEK_PREV;
	if (nKey1 == KEY_NEXT)
		return KEY_SEEK_NEXT;
	
	while(GetKeySub() != KEY_NONE);
	
	return KEY_NONE;
}

// ●キー入力
static BYTE GetKeySub(void)
{
	WORD swv = GetSWV();
	// NONE: 1023
	// PREV:  717 
	// NEXT:  618 
	// BAND:  536 
	// POWER:   1
	
	if (swv >= 900)
		return KEY_NONE;
	if (swv > 667)
		return KEY_PREV;
	if (swv > 577)
		return KEY_NEXT;
	if (swv > 100)
		return KEY_BAND;

	return  KEY_POWER;
}

// ●SW電圧値取得
static WORD GetSWV(void)
{
	ADCON0bits.GO = 1;
	while(ADCON0bits.GO_nDONE);
	
	return ADRES;
}

// ●バンド切り替え
static void SwitchBand(void)
{
	SetMute(TRUE);

	s_bBand = !s_bBand;
	SetBand(s_bBand);

	SleepDelay(0b01000); // 256ms
	SetMute(FALSE);
}

// ●前の周波数へ
static void PrevFreq(void)
{
	if (s_bBand)
	{
		s_wFreqFM = GetPrevFreq(s_wFreqsFM, CHAN_NUM_FM, s_wFreqFM);
		SetFreqFM(s_wFreqFM);
	}
	else
	{
		s_wFreqAM = GetPrevFreq(s_wFreqsAM, CHAN_NUM_AM, s_wFreqAM);
		SetFreqAM(s_wFreqAM);
	}
}

// ●次の周波数へ
static void NextFreq(void)
{
	if (s_bBand)
	{
		s_wFreqFM = GetNextFreq(s_wFreqsFM, CHAN_NUM_FM, s_wFreqFM);
		SetFreqFM(s_wFreqFM);
	}
	else
	{
		s_wFreqAM = GetNextFreq(s_wFreqsAM, CHAN_NUM_AM, s_wFreqAM);
		SetFreqAM(s_wFreqAM);
	}
}

// ●前の周波数を取得
static WORD GetPrevFreq(WORD* pFreq, INT8 nCount, WORD wCur)
{
	int i;
	for(i = nCount - 1; i >= 0; i--)
	{
		if (pFreq[i] < wCur)
			return pFreq[i];
	}
	return pFreq[nCount - 1];
}

// ●次の周波数を取得
static WORD GetNextFreq(WORD* pFreq, INT8 nCount, WORD wCur)
{
	int i;
	for(i = 0; i < nCount; i++)
	{
		if (pFreq[i] > wCur)
			return pFreq[i];
	}
	return pFreq[0];
}

// ●シーク
static void Seek(BOOL bUp)
{
	if (s_bBand)
		s_wFreqFM = SeekFM(bUp);
	else
		s_wFreqAM = SeekAM(bUp);
}

//-----------------------------------------------------------------

// ●受信開始終了
static void EnableRecv(BOOL bEnable)
{
	if (bEnable)
	{
		// Power Up Enable
		WriteI2C(0x02, 0xB003); // Mute, Mono, Bass, Reset, Enable
		SleepDelay(0b00000); // 1ms
		WriteI2C(0x02, 0xB001); // Mute, Mono, Bass, Enable

		WORD wData = ReadI2C(0x04);
		wData |= 0x0800;
		WriteI2C(0x04, wData); // De-emphasis. 50us
		
		// ここで待たないと次の選局が正常に行われない
		SleepDelay(0b01001); // 512ms
		SleepDelay(0b01000); // 256ms
		
		// ステレオ設定
		Stereo(s_bPH);
		
		// バンド設定
		SetBand(s_bBand);

		// ボリューム設定
		SetVolume((s_bPH)? s_nVolume : MAX_VOLUME);

		// ミュート解除
		SetMute(FALSE);
	}
	else
	{
		// Power Up Disable
		WriteI2C(0x02, 0x8000); // Disable
	}
}

// ●ステレオ設定
static void Stereo(BOOL bStereo)
{
	WORD wData = ReadI2C(0x02);
	wData = (bStereo)? (wData & ~0x2000) : (wData | 0x2000);
	WriteI2C(0x02, wData);
}

// ●バンド設定
static void SetBand(BOOL bFM)
{
	if (bFM)
	{
		WriteI2C(0x40, 0x0000); // FM mode
		SetFreqFM(s_wFreqFM);
	}
	else
	{
		WriteI2C(0x40, 0x0002); // AM/LW mode
		SetFreqAM(s_wFreqAM);
	}
}

// ●周波数設定(FM)
static void SetFreqFM(WORD wFreq)
{
	WORD wData;
	wData = (wFreq - 760); // CHAN
	wData <<= 6;
	wData |= 0x0010; // TUNE
	wData |= 0x0008; // BAND, SPACE

	WriteI2C(0x03, wData);

	 // STC待ち
	do {
		SleepDelay(0b00011); // 8ms
		wData = ReadI2C(0x0A);
	}
	while((wData & 0x4000) == 0);
}

// ●周波数設定(AM)
static void SetFreqAM(WORD wFreq)
{
	WriteI2C(0x72, wFreq);

	 // STC待ち
	WORD wData;
	do {
		SleepDelay(0b00011); // 8ms
		wData = ReadI2C(0x0A);
	}
	while((wData & 0x4000) == 0);
}

// ●シーク(FM)
static WORD SeekFM(BOOL bUp)
{
	// ワンショットシーク
	WORD wData = ReadI2C(0x02);
	wData |= 0x0100; // SEEK
	wData &= ~0x0080; // SKMODE
	wData = (bUp)? (wData | 0x0200) : (wData & ~0x0200); // SEEKUP
	WriteI2C(0x02, wData);

	// STC待ち
	do {
		SleepDelay(0b00011); // 8ms
		wData = ReadI2C(0x0A);
	}
	while((wData & 0x4000) == 0);

	// 現在の周波数を取得
	wData = ReadI2C(0x0A);
	WORD wFreq = (wData & 0x02FF) + 760;
	return wFreq;
}

// ●シーク(AM)
static WORD SeekAM(BOOL bUp)
{
	WriteI2C(0x71, 9); // 9KHz step
	WriteI2C(0x73, 522); // 522KHz bottom
	WriteI2C(0x74, 1620); // 1620KHz top

	// ワンショットシーク
	WORD wData = ReadI2C(0x02);
	wData |= 0x0100; // SEEK
	wData &= ~0x0080; // SKMODE
	wData = (bUp)? (wData | 0x0200) : (wData & ~0x0200); // SEEKUP
	WriteI2C(0x02, wData);

	// STC待ち
	do {
		SleepDelay(0b00011); // 8ms
		wData = ReadI2C(0x0A);
	}
	while((wData & 0x4000) == 0);

	// 現在の周波数を取得
	WORD wFreq = ReadI2C(0x7A);
	return wFreq;
}

// ●ボリューム設定
static void SetVolume(BYTE bVol)
{
	WORD wData = ReadI2C(0x05);
	wData = (wData & 0xFFF0) | bVol;

	WriteI2C(0x05, wData);
}

// ●ミュート設定
static void SetMute(BYTE bMute)
{
	WORD wData = ReadI2C(0x02);
	wData = (bMute)? (wData & 0xBFFF) : (wData | 0x4000);

	WriteI2C(0x02, wData);
}

//-----------------------------------------------------------------

// ●I2C送信(RDA5830専用)
static void WriteI2C(BYTE nReg, WORD wData)
{
	// Start conditon
	PIR1bits.SSP1IF = 0;
	SSP1CON2bits.SEN = 1;
	while (!PIR1bits.SSP1IF);

	// 7bit address & W flag
	PIR1bits.SSP1IF = 0;
	SSP1BUF = (ADDR_RDA5830 << 1) | 0x00;
	while (!PIR1bits.SSP1IF);

	// reg address
	PIR1bits.SSP1IF = 0;
	SSP1BUF = nReg;
	while (!PIR1bits.SSP1IF);
	
	// data(H)
	PIR1bits.SSP1IF = 0;
	SSP1BUF = (BYTE)(wData >> 8);
	while (!PIR1bits.SSP1IF);

	// data(L)
	PIR1bits.SSP1IF = 0;
	SSP1BUF = (BYTE)wData;
	while (!PIR1bits.SSP1IF);

	// Stop condition
	PIR1bits.SSP1IF = 0;
	SSP1CON2bits.PEN = 1;
	while (!PIR1bits.SSP1IF);
}

// ●I2C受信(RDA5830専用)
static WORD ReadI2C(BYTE nReg)
{
	// Start conditon
	PIR1bits.SSP1IF = 0;
	SSP1CON2bits.SEN = 1;
	while (!PIR1bits.SSP1IF);

	// 7bit address & W flag
	PIR1bits.SSP1IF = 0;
	SSP1BUF = (ADDR_RDA5830 << 1) | 0x00;
	while (!PIR1bits.SSP1IF);

	// reg address
	PIR1bits.SSP1IF = 0;
	SSP1BUF = nReg;
	while (!PIR1bits.SSP1IF);

	// Start conditon
	PIR1bits.SSP1IF = 0;
	SSP1CON2bits.SEN = 1;
	while (!PIR1bits.SSP1IF);

	// 7bit address & R flag
	PIR1bits.SSP1IF = 0;
	SSP1BUF = (ADDR_RDA5830 << 1) | 0x01;
	while (!PIR1bits.SSP1IF);

	// data
	BYTE Data[2];
	for (BYTE i = 0; i < 2; i++)
	{
		PIR1bits.SSP1IF = 0;
		SSP1CON2bits.RCEN = 1;
		while (!PIR1bits.SSP1IF);
		Data[i] = SSP1BUF;

		// ACK or NACK
		PIR1bits.SSP1IF = 0;
		SSP1CON2bits.ACKDT = i;
		SSP1CON2bits.ACKEN = 1;
		while (!PIR1bits.SSP1IF);
	}

	// Stop condition
	PIR1bits.SSP1IF = 0;
	SSP1CON2bits.PEN = 1;
	while (!PIR1bits.SSP1IF);
	
	return (Data[0] << 8) | Data[1];
}

//-----------------------------------------------------------------

// ●スリープして待つ
static void SleepDelay(WORD uWdtPs)
{
	/*
	10010  = 1:8388608 (223) (Interval 256s nominal)
	10001  = 1:4194304 (222) (Interval 128s nominal)
	10000  = 1:2097152 (221) (Interval 64s nominal)
	01111  = 1:1048576 (220) (Interval 32s nominal)
	01110  = 1:524288 (219) (Interval 16s nominal)
	01101  = 1:262144 (218) (Interval 8s nominal)
	01100  = 1:131072 (217) (Interval 4s nominal)
	01011  = 1:65536  (Interval 2s nominal)
	01010  = 1:32768 (Interval 1s nominal)
	01001  = 1:16384 (Interval 512 ms nominal)
	01000  = 1:8192 (Interval 256 ms nominal)
	00111  = 1:4096 (Interval 128 ms nominal)
	00110  = 1:2048 (Interval 64 ms nominal)
	00101  = 1:1024 (Interval 32 ms nominal)
	00100  = 1:512 (Interval 16 ms nominal)
	00011  = 1:256 (Interval 8 ms nominal)
	00010  = 1:128 (Interval 4 ms nominal)
	00001  = 1:64 (Interval 2 ms nominal)
	00000  = 1:32 (Interval 1 ms nominal)
	*/
	WDTCONbits.WDTPS = uWdtPs;
	SWDTEN = 1;
	CLRWDT();
	SLEEP();
	SWDTEN = 0;
}

//-----------------------------------------------------------------

// ●割込み
void __interrupt() Interrupt(void)
{
	if (IOCIF)
	{
		IOCCFbits.IOCCF3 = 0;
	}
}
